
import { Service } from "../service"
import { MemoryLockService } from "./memoryLockService"
import { MongodbService } from "./mongodbService"
import * as mongodb from "mongodb"
import * as Leaf from "leaf-ts"
import * as UUID from "uuid"

export type Callback<T = any, E extends Error = Error> = (e?: E, result?: T) => void
export const Errors = Leaf.ErrorDoc.build({
    UnknownError: {},
    NotFound: {},
    InvalidParameter: {},
})

export class IncrementalIdService extends Service {
    name = "IncrementalIdService"
    dependencies = ["MongodbService"]
    services: {
        MongodbService: MongodbService
    }

    constructor() {
        super()
    }
    protected collectionName = "IncrementalIdService"
    protected incrementalIdCollection: mongodb.Collection
    public readonly defaultOffset = 1
    initialize() {
        this.incrementalIdCollection = this.services.MongodbService.db.collection(this.collectionName)
        this.incrementalIdCollection.createIndex({
            name: 1
        }, {
                unique: true
            })
    }
    ensure(option: { name: string, offset: number }, callback: Callback<void>) {
        if (!option.name) {
            callback(new Errors.InvalidParameter)
            return
        }
        this.incrementalIdCollection.insertOne({
            name: option.name,
            offset: option.offset | this.defaultOffset
        }, (err) => {
            if (err && err["code"] !== 11000) {
                callback(new Errors.UnknownError("Unknown error when insert base document", { via: err }))
                return
            }
            callback(null)
        })
    }
    next(option: { name: string }, callback: Callback<number>) {
        if (!option.name) {
            callback(new Errors.InvalidParameter)
            return
        }
        this.incrementalIdCollection.findOneAndUpdate({
            name: option.name
        }, {
                offset: {
                    $inc: 1
                }
            }, (err, result) => {
                callback(result.value.offset)
            })
    }
}
